#include <errno.h>
#include <fcntl.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <time.h>
#include <unistd.h>

#define THE_SIZE 0x20000000000
// also known as 2 TiB

void sin(const char *err)
{
	fprintf(stderr, "fatal: sinful system%s%s\n", err ? ": " : "",
	        err ? err : "");
	exit(1);
}

typedef struct {
	uint8_t a : 4, b : 4;
} bitbyte_t;

typedef struct {
	uint64_t total;
	uint64_t off_8;
	uint64_t off_16;
	uint64_t off_32;
	uint64_t off_64;
	uint64_t data;
} data_t;

static data_t *data;

void *align(void *ptr, int ligma)
{
	return (void *)(((size_t)ptr + (ligma - 1)) & -ligma);
}

uint64_t get(uint64_t i)
{
	if (i >= data->total) {
		sin("sinful memory access");
	}

	char *ptr = (char *)&data->data;
	uint64_t off = 0;

	if ((data->off_8 - 1) <= i) {
		ptr += (data->off_8 - 1 - off) / 2 +
		       (data->off_8 - 1 - off) % 2;
		ptr = align(ptr, 1);
		off += data->off_8 - 1;
	} else {
		bitbyte_t *v = &((bitbyte_t *)ptr)[(i - off) / 2];
		return (i - off) % 2 ? v->b + 1 : v->a + 1;
	}

	if ((data->off_16 - 1) <= i) {
		ptr += 1 * (data->off_16 - 1 - off);
		ptr = align(ptr, 2);
		off += data->off_16 - 1;
	} else {
		return ((uint8_t *)ptr)[i - off] + 1;
	}

	if ((data->off_32 - 1) <= i) {
		ptr += 2 * (data->off_32 - 1 - off);
		ptr = align(ptr, 4);
		off += data->off_32 - 1;
	} else {
		return ((uint16_t *)ptr)[i - off] + 1;
	}

	if ((data->off_64 - 1) <= i) {
		ptr += 4 * (data->off_64 - 1 - off);
		ptr = align(ptr, 8);
		off += data->off_64 - 1;
	} else {
		return ((uint32_t *)ptr)[i - off] + 1;
	}

	return ((uint64_t *)ptr)[i - off] + 1;
}

void append(uint64_t val)
{
	uint64_t i = data->total++;

	if (val < 1) {
		sin("sinful data");
	}
	val--;

	if (val >= (1ULL << 4) && !data->off_8) {
		data->off_8 = i + 1;
	}

	if (val >= (1ULL << 8) && !data->off_16) {
		data->off_16 = i + 1;
	}

	if (val >= (1ULL << 16) && !data->off_32) {
		data->off_32 = i + 1;
	}

	if (val >= (1ULL << 32) && !data->off_64) {
		data->off_64 = i + 1;
	}

	char *ptr = (char *)&data->data;
	uint64_t off = 0;

	if ((data->off_8 - 1) <= i) {
		ptr += (data->off_8 - 1 - off) / 2 +
		       (data->off_8 - 1 - off) % 2;
		ptr = align(ptr, 1);
		off += data->off_8 - 1;
	} else {
		bitbyte_t *v = &((bitbyte_t *)ptr)[(i - off) / 2];
		(i - off) % 2 ? (v->b = val) : (v->a = val);
		return;
	}

	if ((data->off_16 - 1) <= i) {
		ptr += 1 * (data->off_16 - 1 - off);
		ptr = align(ptr, 2);
		off += data->off_16 - 1;
	} else {
		((uint8_t *)ptr)[i - off] = val;
		return;
	}

	if ((data->off_32 - 1) <= i) {
		ptr += 2 * (data->off_32 - 1 - off);
		ptr = align(ptr, 4);
		off += data->off_32 - 1;
	} else {
		((uint16_t *)ptr)[i - off] = val;
		return;
	}

	if ((data->off_64 - 1) <= i) {
		ptr += 4 * (data->off_64 - 1 - off);
		ptr = align(ptr, 8);
		off += data->off_64 - 1;
	} else {
		((uint32_t *)ptr)[i - off] = val;
		return;
	}

	((uint64_t *)ptr)[i - off] = val;
}

int generator(const char *path)
{
	int fd = open(path, O_RDWR | O_CREAT, 0644);
	if (fd == -1)
		sin(strerror(errno));
	if (truncate(path, THE_SIZE))
		sin(strerror(errno));

	data = mmap(NULL, THE_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED,
	            fd, 0);

	if (data == MAP_FAILED)
		sin(strerror(errno));

	int column = 0;

	int human_readable = isatty(fileno(stdout));

	for (uint64_t i = 0; i < data->total; i++) {

		column += printf("%llu", (unsigned long long)get(i));

		if (!human_readable) {
			column = 100;
		}

		if (column > 72) {
			printf("\n");
			column = 0;
		} else {
			column += printf(" ");
		}
	}

	struct timespec ts_begin;
	clock_gettime(CLOCK_MONOTONIC, &ts_begin);

	for (;;) {

		uint64_t max_rep = 1;

		for (uint64_t stride = 1;
		     stride <= data->total / (max_rep + 1); stride++) {

			uint64_t rep;

			for (rep = 1; (rep + 1) <= data->total / stride;
			     rep++) {
				for (uint64_t i = 0; i < stride; i++) {
					if (get(data->total - 1 - i -
					        rep * stride) !=
					    get(data->total - 1 - i -
					        (rep - 1) * stride)) {
						goto repeat_out;
					}
				}
			}

		repeat_out:

			if (max_rep < rep) {
				max_rep = rep;
			}
		}

		column += printf("%llu", (unsigned long long)max_rep);

		if (!human_readable) {
			column = 100;
		}

		if (column > 72) {
			printf("\n");
			column = 0;
		} else {
			column += printf(" ");
		}

		struct timespec ts_end;
		clock_gettime(CLOCK_MONOTONIC, &ts_end);

		if ((ts_end.tv_sec * 1000 + ts_end.tv_nsec / 1000000) -
		        (ts_begin.tv_sec * 1000 +
		         ts_end.tv_nsec / 1000000) >
		    50) {
			fflush(stdout);
			clock_gettime(CLOCK_MONOTONIC, &ts_begin);
		}

		append(max_rep);
	}

	return 0;
}

int main()
{
	const char *cfg_home = getenv("XDG_CONFIG_HOME");

	if (!cfg_home) {

		const char *home = getenv("HOME");
		if (!home) {
			return generator("/tmp/cg666.bin");
		}

		char path[strlen(home) +
		          sizeof("/.config/cg666/data.bin")];
		path[0] = '\0';

		strcat(path, home);
		strcat(path, "/.config/cg666");

		pid_t pid;
		if (!fork()) {
			if (execl("/usr/bin/env", "/usr/bin/env",
			          "mkdir", "-p", "--", path, NULL))
				sin(strerror(errno));
			return 0;
		}
		int status;
		if (waitpid(pid, &status, 0) == -1) {
			sin(strerror(errno));
		} else if (status != 0) {
			sin("mkdir is a sinner");
		}

		strcat(path, "/data.bin");
		return generator(path);

	} else {

		char path[strlen(cfg_home) + sizeof("/cg666/data.bin")];
		path[0] = '\0';

		strcat(path, cfg_home);
		strcat(path, "/cg666");

		pid_t pid;
		if (!fork()) {
			if (execl("/usr/bin/env", "/usr/bin/env",
			          "mkdir", "-p", "--", path, NULL))
				sin(strerror(errno));
			return 0;
		}
		int status;
		if (waitpid(pid, &status, 0) == -1) {
			sin(strerror(errno));
		} else if (status != 0) {
			sin("mkdir is a sinner");
		}

		strcat(path, "/data.bin");
		return generator(path);
	}
}
